##################################################################################
#   Version 8.21
#   
#   File: clientUtils.tcl
#
#   Copyright   IXIA
#   All Rights Reserved.
#
#    Revision Log:
#    Date           Author                 Comments
#    -----------    -------------------    --------------------------------------------
#    10/25/2000     ds                      initial release
#
# Description:  This file contains utilities for the client side of the Tcl proxy
#
##################################################################################

if {[isUNIX]} {
    package req ssh
}

########################################################################################
# Procedure:    clientOpen
#
# Description:  Open a connection to the ixTclServer.
#                
# Input:        server:        host name of server
#               port:        port id of service
#
# Output:       socket handle
#
########################################################################################
proc clientOpen {host port} \
{
    if [catch {socket $host $port} socketId] {
        set socketId {}
    }

    return $socketId
}


########################################################################################
# Procedure:    clientOpenSSH
#
# Description:  Open a connection to the ixTclServer.
#                
# Input:        server:        host name of server
#               port:        port id of service
#
# Output:       serverIP
#
########################################################################################
proc clientOpenSSH {host port} \
{
    variable privateKey
    # Get the private key location
    set absolutePath [file dirname [dict get [info frame [info frame]] file]]
    set privateKey "$absolutePath/../id_rsa"
    tclServer::setTclServerType 0
    set result {}
    if [catch {
        set username $::tcl_platform(user)
        set defaultFolder "/tmp/$username/.ssh"
        set homeFolder $defaultFolder
        if {[info exists ::env(HOME)]} {
            set homeFolder "$::env(HOME)/.ssh"
        }
        
        set userPrivateKey "$homeFolder/ixpit_$username\_id_rsa"
        if [catch {
            setupPrivateKey $homeFolder $privateKey $userPrivateKey
            } ] {
                set userPrivateKey "$defaultFolder/ixpit_$username\_id_rsa"
                setupPrivateKey $defaultFolder $privateKey $userPrivateKey
            }

        # With VM chassis is common to redeploy a chassis with the same IP, because of this we have to ignore the host fingerprint
        ::ssh::connect -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -p $port -i $userPrivateKey -T ixtcl@$host
        set ixosVersion {}

        if {[info exists ::env(IXOS_VERSION)]} {
            set ixosVersion $::env(IXOS_VERSION)
            puts "version $ixosVersion"
        } else {
            if [catch {clientSend $host "exec /opt/ixia/appinfo/reg.py --get-reg {hkey_local_machine\\software\\ixia communications\\ixserver\\settings\\lastrunixosversion}"} r] {
                errorMsg $r
                set result {}
            } else {
                set fullIxosVersion $r
                if {$fullIxosVersion == "None"} {set ixosVersion {}} else {
                set ixosVersion [lindex [split [lindex [split $fullIxosVersion {"("}] 1] {")"}] 0]
                }
            }
        }
        if {$ixosVersion == {}} {
            errorMsg "Error: No IxOS version found on the tcl server!"
            set result {}
        } else {
            if [catch {clientSend $host "source {/opt/ixia/ixos/$ixosVersion/IxiaWish.tcl}"} r] {
                errorMsg $r
                set result {}
            } else {
                puts $r
                set result $host
            }
        }
    } errmsg] {
        errorMsg "Error: $errmsg"
        set result {}
        return $result
    }
    if {$result == {}} {
        clientClose $host
    }

    return $result
}

########################################################################################
# Procedure:    clientClose
#
# Description:  Close connection from the client side.
#                
# Input:        serverId:    client-side socket or server IP
#
# Output:        0 if successful
#                1 if error while attempting to close socket
#
########################################################################################
proc clientClose {serverId} \
{
    set retCode 0
    if {[tclServer::isWindowsTclServer]} {
        if [catch {close $serverId}] {
            set retCode 1
        }
    } else {
        if [catch {::ssh::disconnect $serverId}] {
            set retCode 1
        }
    }
    return $retCode
}


########################################################################################
# Procedure:    clientSend
#
# Description:  Send a command from the client side.
#                
# Input:        serverId:    client-side socket or serverIP
#                args:        TCL command to evaluate
#
# Returns:        Success:    TCL Return result
#               Failure:    {}
#
# Remarks:  TCL procs can embed i/o.
#
########################################################################################
proc clientSend {serverId args} \
{
    set retCode   1
    set retResult 0

    #
    # send data over the socket
    #
    set buf [lindex $args 0]

    #   
    #   Return buffer is formatted as follows:
    #
    #   0 to final lf/cr -> Tcl Standard Output
    #   final character  -> Tcl Return Code (TCL_OK, TCL_ERROR)
    #
    if [tclServer::isWindowsTclServer] {
        if [catch { puts $serverId $buf ; 
                    flush $serverId ; 
                    #
                    # read the reply
                    #  the reply may have the format 
                    #       sOutput/r/nsTclResult/r/n    -- i/o output followed by TCL result
                    #                                                 /r/n -delimited
                    #       sTclresult/r/nsTclResultCode -- simple TCL result
                    #       null                         -- no TCL result available
                    
                    vwait tclServer::buffer
                    set retBuffer $tclServer::buffer
                    
                    set indexOfLastCrlf [string last "\r\n" $retBuffer]
                    if {$indexOfLastCrlf != -1 } {
                        set lenBuffer [string length $retBuffer]
                        set indexOfPenultimateCrlf [string last "\r\n" [string range $retBuffer 0 [expr $indexOfLastCrlf -1]]]
                        if {$indexOfPenultimateCrlf != -1 } {
                            set retResult [string range $retBuffer [expr $indexOfPenultimateCrlf + 2] $lenBuffer]
                        } else {
                            set length    [string length $retBuffer]
                            set retCode   [string index $retBuffer [incr length -1]]
                            set retResult [string range $retBuffer 0 [incr length -1]]
                        }
                    } else {
                        set length    [string length $retBuffer]
                        set retCode   [string index $retBuffer [incr length -1]]
                        set retResult [string range $retBuffer 0 [incr length -1]]
                    }
                } 
        ] {
            errorMsg $::errorInfo
            tclServer::disconnectTclServer
            set retResult 1
        }
    } else {
        set script [format "puts \"\[puts \"\n\[catch {%s} result\]\"\];\[puts \$result\];DONE 0\"" $buf]
        ::ssh::push $serverId $script
        ::ssh::send $serverId;
        # the result will be a list: {retCode, retOutput}
        set res [list]
        while {[::ssh::pop_line $serverId line] >= 0} {
            if {[string first "% " $line] == 0} {
               continue
            }

            if {[regexp {^;;DONE\s+(\d+)} $line match code]} {
                break;
            }
            if {$line != {}} {
                lappend res $line
            }
        }

        set len [llength $res]
        set retResult ""
        if {$len == 0} {
            set retCode 0
            set retResult ""
        }
        if {$len == 1} {
            set retCode [lindex $res 0]
            set retResult ""
        }
        if {$len == 2} {
            set retCode [lindex $res 0]
            set retResult [lindex $res 1]
        }
        if {$len > 2} {
            set retCode [lindex $res 0]
            for {set i 1} {$i < $len} {incr i} {
                set retResult [concat $retResult [lindex $res $i] "\n"]
            }
        }
    }
    #
    # Force an error if the command returned TCL_ERROR.
    #   Can't use the constant TCL_ERROR here since it is not defined at
    #   this point in execution.
    #
    if {$retCode == 1} {
        set retCommand [list error $retResult $retResult]
    } else {
        set retCommand [list return $retResult]
    }

    eval $retCommand
}

########################################################################################
# Procedure:    remoteDefine
#
# Description:     Create a proc to proxy over a ixTclHal command.
#                
# Input:        commandList
#
# Output:        0 if successful
#                1 if error
#
########################################################################################
proc remoteDefine { commandList } \
{
    foreach procName $commandList {
        set procNameOld ${procName}Old
        if { [info commands $procName] != "" } {
            if { [info commands $procNameOld] == "" } {
                rename $procName $procNameOld
            }
        }
        eval [format   "proc %s {args} \
                        {\
                            global ixTclSvrHandle; \
                            if {!\[tclServer::isTclServerConnected\]} { \
                                if \[catch { eval \"%s \$args\" } result \] {error \$result \$result }; \
                                return \$result }; \
                            if \[catch { eval \"clientSend \$ixTclSvrHandle {%s \$args}\" } result\] {error \$result \$result}; \
                            if  {\$result != \"\"} {if \[catch { eval \"clientSend \$ixTclSvrHandle {set ixErrorInfo}\" } ::ixErrorInfo] {error \$::ixErrorInfo \$::ixErrorInfo}}; \
                            return \$result \
                        }" \
        $procName $procNameOld $procName]
    }
    return 0
}

########################################################################################
# Procedure:    getConstantsValue
#
# Description:     Get the list of constants from ixTclServer.
#                
# Input:        serverSocket
#
# Output:        0 if successful
#                1 if error
#
########################################################################################
proc getConstantsValue {serverSocket} \
{
    set retCode 0

    set constList  [clientSend $serverSocket {array get ixConstants}]
    if {[llength $constList] > 0} {
        foreach {constName constVal} $constList {
            global $constName
            # The catch prevents the error message from flowing back on windows
            catch {set $constName $constVal}
        }
    } else {
        set retCode 1
    }

    return $retCode
}


########################################################################################
# Procedure:    ixMasterSet
#
# Description:     No clue what this is here for, but I'm leaving it for backwards compatibility
#
# Input:
#
########################################################################################
proc ixMasterSet {name element op} \
{
    upvar ${name}($element) master
    tclServer::connectToTclServer $master errMsg
}

########################################################################################
# Procedure:    redefineCommand
#
# Description:     Redefine the specified command to make its methods import and export work 
#               for UNIX client.
#
# Input:        command - name of the command to be redefined.
#
########################################################################################
proc redefineCommand {command} \
{
    set commandOld ${command}Oldish

    if { [info command $command] != "" } {
        if { [info command $commandOld] == "" } {
            rename $command $commandOld
        }
    }
    if {[tclServer::isWindowsTclServer]} {
        eval [format "proc %s {args} \
              { \
                  set cmdLine %s ; \
                  set path \[file dirname \[lindex \$args 1\]\] ;\
                  set fileName \[file tail \[lindex \$args 1\]\] ;\

                  if { \$path == \".\" } { \
                    append cmdLine \" \$args\"; set path \$fileName } else { \
                    append cmdLine \" \[lindex \$args 0\] \$fileName \[lindex \$args 2\] \[lindex \$args 3\] \[lindex \$args 4\] \" }\

                  switch \[lindex \$args 0\] { \
                      import { \
                          doFileTransfer \"put\" \[lindex \$args 1\] \$fileName ; \
                          eval \$cmdLine; \
                      } \
                      export { \
                          set retCode \[eval \$cmdLine\]; \
                          if \{\$retCode == 0\} \{ \
                              doFileTransfer \"get\" \$fileName \[lindex \$args 1\]; \
                              fileTransferClient::deleteFile \[tclServer::getTclServerName\] 4500 \$fileName \

                          \} else \{ \
                              return \$retCode; \
                           \} \
                      } \
                      default { \
                          eval \$cmdLine; \
                      } \
                   } \
               }" \
        $command $commandOld]
    } else {
        eval [format "proc %s {args} \
            { \
                set cmdLine %s ; \
                set path \[file dirname \[lindex \$args 1\]\] ;\
                set fileName \[file tail \[lindex \$args 1\]\] ;\ 
                switch \[lindex \$args 0\] { \
                    import { \
                        doFileTransfer \"put\" \[lindex \$args 1\] \$fileName ; \
                        append cmdLine \" \[lindex \$args 0\] \[getFtpFolder\]\$fileName \[lindex \$args 2\] \[lindex \$args 3\] \[lindex \$args 4\] \";\
                        set retCode \[eval \$cmdLine\]; \
                        fileTransferClient::deleteFile \[tclServer::getTclServerName\] 4500 \$fileName; \
                        return \$retCode; \
                    } \
                    export { \
                        append cmdLine \" \[lindex \$args 0\] \[getFtpFolder\]\$fileName \[lindex \$args 2\] \[lindex \$args 3\] \[lindex \$args 4\] \";\
                        set retCode \[eval \$cmdLine\]; \
                        if \{\$retCode == 0\} \{ \
                            doFileTransfer \"get\" \$fileName \[lindex \$args 1\]; \
                            fileTransferClient::deleteFile \[tclServer::getTclServerName\] 4500 \$fileName \                       
                              
                        \} else \{ \
                            return \$retCode; \
                        \} \
                    } \
                    default { \
                        append cmdLine \" \$args\";\
                        eval \$cmdLine; \
                    } \
                } \
            }" \
        $command $commandOld]
    }
}

########################################################################################
# Procedure:    doFileTransfer
#
# Description:     Transfer files between client and chassis.
#
# Input:        action    - Has to be either put or get.
#                            
# Unix doesn't need the full path of the file, that is why it is based on the direction 
# if it is put or get, we have to pass in two filenames, one for source the other for
# destination  
#               filename1  - Name of the file to be transfered.
#               filename2  - Name of the file to be transfered. 
#
########################################################################################
proc doFileTransfer {action filename1 filename2 {port 4500}} \
{
    set retCode 1

    if [tclServer::isTclServerConnected] {
        set retCode [fileTransferClient::${action}File [tclServer::getTclServerName] $port "$filename1" "$filename2"]
    }

    return $retCode
}


proc getFtpFolder {} \
{
    variable fullFtpFolder "/ftp/virtual/ixia/tclftp/"
    return $fullFtpFolder
}


proc setupPrivateKey {folder privateKey userPrivateKey} \
{
    if {[file exists $userPrivateKey] != 1} {
        exec "mkdir" "-p" "$folder/"
        file copy $privateKey $userPrivateKey
    }
    exec "chmod" "700" "$folder"
    exec "chmod" "600" "$userPrivateKey"
}